# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import inspect
import logging

from deluca.core import JaxObject

AgentRegistry = {}

logger = logging.getLogger(__name__)

# TODO
# - Support hierarchical agents


def make_agent(cls, *args, **kwargs):
    if cls not in AgentRegistry:
        raise ValueError(f"Agent `{cls}` not found.")
    return AgentRegistry[cls](*args, **kwargs)


class Agent(JaxObject):
    @classmethod
    def __init_subclass__(cls, *args, **kwargs):
        """For avoiding a decorator for each subclass"""
        super().__init_subclass__(*args, **kwargs)
        if cls.__name__ not in AgentRegistry and not inspect.isabstract(cls):
            AgentRegistry[cls.__name__] = cls

    def __call__(self, observation):
        raise NotImplementedError()

    def reset(self):
        logger.warn(f"`reset` for agent `{self.name}` not implemented")

    def feed(self, reward):
        logger.warn(f"`feed` for agent `{self.name}` not implemented")



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""deluca.agents._ilqr

Todos:
    - Leverage LQR agent
    - Remove ILQRInner
"""
import time

import jax.numpy as jnp

from deluca.agents.core import Agent


def iLQR_loop(env, U_initial, T, alpha=1.0, log=None):
    U, X, k, K = U_initial, None, None, None
    X, U, c = rollout(env, U, alpha=alpha)
    print("initial cost:" + str(c))
    r = 1
    # while t <= T:
    for t in range(T):
        F, C = f_at_x(env, X, U)
        k, K = LQR(F, C)
        for alphaC in alpha * 1.1 * 1.1 ** (-jnp.arange(10) ** 2):
            # t += 1
            r += 1
            XC, UC, cC = rollout(env, U, k, K, X, alphaC)

            if cC <= c:
                X, U, c, alpha = XC, UC, cC, alphaC
                # print("iLQR : t = %d, c = %f" % (t, min(c, cC)))
                print(f"(iLQR): t = {t}, r = {r}, c = {c}, alpha = {alpha}")
                if log is not None:
                    # log.append((t, float(min(c, cC)), env.nsamples))
                    log.append(f"(iLQR): t = {t}, r = {r}, c = {c}, alpha = {alpha}")
                break
            # if t == T:
            #     return X, U, k, K, c
    return X, U, k, K, c


def LQR(F, C):
    H, d = len(C), C[0][0].shape[0]
    K, k = (
        [None for _ in range(H)],
        [None for _ in range(H)],
    )
    V_x, V_xx = jnp.zeros(d), jnp.zeros((d, d))
    for h in range(H - 1, -1, -1):
        f_x, f_u = F[h]
        c_x, c_u, c_xx, c_uu = C[h]
        Q_x, Q_u = c_x + f_x.T @ V_x, c_u + f_u.T @ V_x
        Q_xx, Q_ux, Q_uu = (
            c_xx + f_x.T @ V_xx @ f_x,
            f_u.T @ V_xx @ f_x,
            c_uu + f_u.T @ V_xx @ f_u,
        )
        # line may be numerically unstable if Q_uu is the 0 matrix
        K[h], k[h] = -jnp.linalg.inv(Q_uu) @ Q_ux, -jnp.linalg.inv(Q_uu) @ Q_u
        V_x = Q_x - K[h].T @ Q_uu @ k[h]
        V_xx = Q_xx - K[h].T @ Q_uu @ K[h]
        V_xx = (V_xx + V_xx.T) / 2
    return k, K


def rollout(env, U_old, k=None, K=None, X_old=None, alpha=1.0, render=False):
    X, U = [None for _ in range(env.H + 1)], [None for _ in range(env.H)]
    X[0], cost = env.reset(), 0.0
    for h in range(env.H):
        if k is None:
            U[h] = U_old[h]
        else:
            U[h] = U_old[h] + alpha * k[h] + K[h] @ (X[h] - X_old[h])
        X[h + 1], instant_cost, _, _ = env.step(U[h])
        cost += instant_cost
        if render:
            env.render()
            time.sleep(0.01)
    return X, U, cost


def f_at_x(env, X, U):
    F, C = [None for _ in range(env.H)], [None for _ in range(env.H)]
    for h in range(env.H):
        F[h] = env.f_x(X[h], U[h]), env.f_u(X[h], U[h])
        C[h] = (
            env.c_x(X[h], U[h]),
            env.c_u(X[h], U[h]),
            env.c_xx(X[h], U[h]),
            env.c_uu(X[h], U[h]),
        )
    return F, C


class ILQRInner:
    def __init__(self, env):
        self.env = env

    def train(self, T, U_init=None, alpha=1.0):
        self.log = []
        if U_init is None:
            U_init = [jnp.zeros(self.env.action_dim) for _ in range(self.env.H)]
        X, U, k, K, c = iLQR_loop(self.env, U_init, T, alpha, self.log)
        return X, U, k, K, c, self.log


class ILQR(Agent):
    def __init__(self):
        self.reset()

    def train(self, sim, train_steps, U_init=None, alpha=1.0):
        self.sim = sim
        ilqr_agent = ILQRInner(self.sim)
        X, U, k, K, c, log = ilqr_agent.train(train_steps, U_init, alpha)
        self.U = jnp.array(U)
        self.X = jnp.array(X)
        self.k = jnp.array(k)
        self.K = jnp.array(K)
        return c, log

    def reset(self):
        self.t = 0

    def __call__(self, state):
        action = self.U[self.t]
        self.t += 1

        return action



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import jax
import jax.numpy as jnp

from deluca.envs.lung import BreathWaveform
from deluca.envs.lung.core import Lung


class DelayLung(Lung):
    def __init__(
        self,
        min_volume=1.5,
        R_lung=10,
        C_lung=6,
        delay=25,
        inertia=0.995,
        control_gain=0.02,
        dt=0.03,
        waveform=None,
        reward_fn=None,
    ):
        self.viewer = None
        self.delay = delay
        self.control_gain = control_gain
        self.inertia = inertia
        self.dt = dt
        self.C_lung = C_lung
        self.R_lung = R_lung
        self.min_volume = min_volume
        self.waveform = waveform or BreathWaveform()
        self.r0 = (3 * self.min_volume / (4 * jnp.pi)) ** (1 / 3)

        self.reset()

    def reset(self):
        self.volume = self.min_volume
        self.pipe_pressure = 0
        self.pressure = 0
        self.time = 0.0
        self.target = self.waveform.at(self.time)
        self.in_history = jnp.zeros(self.delay)
        self.out_history = jnp.zeros(self.delay)

        self.state = {
            "volume": self.volume,
            "pressure": self.pressure,
            "pipe_pressure": self.pipe_pressure,
        }
        return self.observation

    @property
    def observation(self):
        return {
            "measured": self.state["pressure"],
            "target": self.target,
            "dt": self.dt,
            "phase": self.waveform.phase(self.time),
        }

    def dynamics(self, state, action):
        """
        state: (volume, pressure)
        action: (u_in, u_out)
        """
        flow = self.state["pressure"] / self.R_lung
        volume = self.state["volume"] + flow * self.dt
        volume = jnp.maximum(volume, self.min_volume)

        r = (3.0 * self.volume / (4.0 * jnp.pi)) ** (1.0 / 3.0)
        lung_pressure = self.C_lung * (1 - (self.r0 / r) ** 6) / (self.r0 ** 2 * r)

        pipe_impulse = jax.lax.cond(
            self.time < self.delay,
            lambda x: 0.0,
            lambda x: self.control_gain * self.in_history[0],
            None,
        )
        peep = jax.lax.cond(
            self.time < self.delay, lambda x: 0.0, lambda x: self.out_history[0], None
        )

        pipe_pressure = self.inertia * state["pipe_pressure"] + pipe_impulse
        pressure = jnp.maximum(0, pipe_pressure - lung_pressure)

        pipe_pressure = jax.lax.cond(peep, lambda x: x * 0.995, lambda x: x, pipe_pressure)

        return {"volume": volume, "pressure": pressure, "pipe_pressure": pipe_pressure}

    def step(self, action):
        u_in, u_out = action
        u_in = jax.lax.cond(u_in > 0.0, lambda x: x, lambda x: 0.0, u_in)

        self.in_history = jnp.roll(self.in_history, shift=1)
        self.in_history = self.in_history.at[0].set(u_in)
        self.out_history = jnp.roll(self.out_history, shift=1)
        self.out_history = self.out_history.at[0].set(u_out)

        self.target = self.waveform.at(self.time)
        reward = -jnp.abs(self.target - self.state["pressure"])

        self.state = self.dynamics(self.state, action)
        self.time += 1

        return self.observation, reward, False, {}



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import jax
import jax.numpy as jnp

from deluca.core import Env
from deluca.core import field


class LDS(Env):
    A: jnp.array = field(jaxed=False)
    B: jnp.array = field(jaxed=False)
    C: jnp.array = field(jaxed=False)
    key: int = field(jax.random.PRNGKey(0), jaxed=False)
    state_size: int = field(1, jaxed=False)
    action_size: int = field(1, jaxed=False)

    def init(self):
        return jax.random.normal(self.key, shape=(self.state_size, 1))

    def __call__(self, state, action):
        new_state = self.A @ state + self.B @ action
        obs = self.C @ new_state

        return new_state, obs



import jax.numpy as jnp

from deluca.agents.core import Agent
from deluca.utils.planning import f_at_x
from deluca.utils.planning import LQR
from deluca.utils.planning import rollout

# from deluca.agents._ilqr import LQR


def iLC_closed(env_true, env_sim, U, T, k=None, K=None, X=None, ref_alpha=1.0, log=None):
    alpha, r = ref_alpha, 1
    X, U, c = rollout(env_true, U, k, K, X)
    print("initial cost:" + str(c))
    for t in range(T):
        _, F, C = f_at_x(env_sim, X, U)
        k, K = LQR(F, C)
        for alphaC in alpha * 1.1 * 1.1 ** (-jnp.arange(10) ** 2):
            r += 1
            XC, UC, cC = rollout(env_true, U, k, K, X, alphaC)
            if cC <= c:
                X, U, c, alpha = XC, UC, cC, alphaC
                print(f"(iLC): t = {t}, r = {r}, c = {c}, alpha = {alpha}")
                if log is not None:
                    log.append(f"(iLC): t = {t}, r = {r}, c = {c}, alpha = {alpha}")
                break
        else:
            return X, U, k, K, c
    return X, U, k, K, c


class ILCInner:
    def __init__(self, env_true, env_sim):
        self.env_true = env_true
        self.env_sim = env_sim

    def train(self, T, U_init=None, k=None, K=None, X=None, ref_alpha=1.0):
        self.log = []
        if U_init is None:
            U_init = [jnp.zeros(self.env_true.action_dim) for _ in range(self.env_true.H)]
        X, U, k, K, c = iLC_closed(
            self.env_true, self.env_sim, U_init, T, k, K, X, ref_alpha, self.log
        )
        return X, U, k, K, c, self.log


class ILC(Agent):
    def __init__(self):
        self.reset()

    def train(
        self, env_true, env_sim, train_steps, U_init=None, k=None, K=None, X=None, ref_alpha=1.0
    ):
        self.env_true = env_true
        self.env_sim = env_sim
        ilqr_agent = ILCInner(self.env_true, self.env_sim)
        X, U, k, K, c, log = ilqr_agent.train(train_steps, U_init, k, K, X, ref_alpha)
        self.U = jnp.array(U)
        self.X = jnp.array(X)
        self.k = jnp.array(k)
        self.K = jnp.array(K)
        return c, log

    def reset(self):
        self.t = 0

    def __call__(self, state):
        action = self.U[self.t]
        self.t += 1

        return action



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""deluca.agents._hinf"""
from numbers import Real
from typing import Tuple

import jax.numpy as jnp
from jax.numpy.linalg import inv

from deluca.agents.core import Agent


class Hinf(Agent):
    """
    Hinf: H-infinity controller (approximately).  Solves a lagrangian min-max
    dynamic program similiar to H-infinity control rescales disturbance to have
    norm 1.
    """

    def __init__(
        self,
        A: jnp.ndarray,
        B: jnp.ndarray,
        T: jnp.ndarray,
        Q: jnp.ndarray = None,
        R: jnp.ndarray = None,
    ) -> None:
        """
        Description: initializes the Hinf agent

        Args:
            A (jnp.ndarray):
            B (jnp.ndarray):
            T (jnp.ndarray):
            Q (jnp.ndarray):
            R (jnp.ndarray):

        Returns:
            None
        """
        d_x, d_u = B.shape

        if Q is None:
            self.Q = jnp.identity(d_x, dtype=jnp.float32)

        if R is None:
            self.R = jnp.identity(d_u, dtype=jnp.float32)

        # self.K, self.W = solve_hinf(A, B, Q, R, T)
        self.t = 0

    def train(self, A, B, T):
        self.K, self.W = solve_hinf(A, B, self.Q, self.R, T)

    def __call__(self, state) -> jnp.ndarray:
        """
        Description: provide an action given a state

        Args:
            state (jnp.ndarray): the error PID must compensate for

        Returns:
            action (jnp.ndarray): action to take
        """
        action = -self.K[self.t] @ state
        self.t += 1
        return action


def solve_hinf(
    A: jnp.ndarray,
    B: jnp.ndarray,
    Q: jnp.ndarray,
    R: jnp.ndarray,
    T: jnp.ndarray,
    gamma_range: Tuple[Real, Real] = (0.1, 10.8),
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """
    Description: solves H-infinity control problem

    Args:
        A (jnp.ndarray):
        B (jnp.ndarray):
        T (jnp.ndarray):
        Q (jnp.ndarray):
        R (jnp.ndarray):

    Returns
        Tuple[jnp.ndarray, jnp.ndarray]:
    """
    gamma_low, gamma_high = gamma_range
    n, m = B.shape
    M = [jnp.zeros(shape=(n, n))] * (T + 1)
    K = [jnp.zeros(shape=(m, n))] * T
    W = [jnp.zeros(shape=(n, 1))] * T

    while gamma_high - gamma_low > 0.01:
        K = [jnp.zeros(shape=(m, n))] * T
        W = [jnp.zeros(shape=(n, 1))] * T

        gamma = 0.5 * (gamma_high + gamma_low)
        M[T] = Q
        for t in range(T - 1, -1, -1):
            Lambda = jnp.eye(n) + (B @ inv(R) @ B.T - gamma ** (-2) * jnp.eye(n)) @ M[t + 1]
            M[t] = Q + A.T @ M[t + 1] @ inv(Lambda) @ A

            K[t] = -inv(R) @ B.T @ M[t + 1] @ inv(Lambda) @ A
            W[t] = (gamma ** (-2)) * M[t + 1] @ inv(Lambda) @ A
            if not is_psd(M[t], gamma):
                gamma_low = gamma
                break
        if gamma_low != gamma:
            gamma_high = gamma
    return K, W


def is_psd(P: jnp.ndarray, gamma: Real) -> bool:
    """
    Description: check if a matrix is positive semi-definite

    Args:
        P (jnp.ndarray): 
        gamma (Real): 

    Returns:
        bool: 
    """
    return jnp.all(jnp.linalg.eigvals(P) < gamma ** 2 + 1e-5)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""deluca.agents._gpc"""
from numbers import Real
from typing import Callable

import jax
import jax.numpy as jnp
from jax import grad
from jax import jit

from deluca.agents.core import Agent
from deluca.utils import Random


def quad_loss(x: jnp.ndarray, u: jnp.ndarray) -> Real:
    """
    Quadratic loss.

    Args:
        x (jnp.ndarray):
        u (jnp.ndarray):

    Returns:
        Real
    """
    return jnp.sum(x.T @ x + u.T @ u)


class DRC(Agent):
    def __init__(
        self,
        A: jnp.ndarray,
        B: jnp.ndarray,
        C: jnp.ndarray = None,
        K: jnp.ndarray = None,
        cost_fn: Callable[[jnp.ndarray, jnp.ndarray], Real] = None,
        m: int = 10,
        h: int = 50,
        lr_scale: Real = 0.03,
        decay: bool = True,
        RM: int = 1000,
        seed: int = 0,
    ) -> None:
        """
        Description: Initialize the dynamics of the model.

        Args:
            A (jnp.ndarray): system dynamics
            B (jnp.ndarray): system dynamics
            C (jnp.ndarray): system dynamics
            Q (jnp.ndarray): cost matrices (i.e. cost = x^TQx + u^TRu)
            R (jnp.ndarray): cost matrices (i.e. cost = x^TQx + u^TRu)
            K (jnp.ndarray): Starting policy (optional). Defaults to LQR gain.
            start_time (int):
            cost_fn (Callable[[jnp.ndarray, jnp.ndarray], Real]):
            H (postive int): history of the controller
            HH (positive int): history of the system
            lr_scale (Real):
            decay (boolean):
            seed (int):
        """

        cost_fn = cost_fn or quad_loss

        self.random = Random(seed)

        d_state, d_action = B.shape  # State & Action Dimensions

        C = jnp.identity(d_state) if C is None else C

        d_obs = C.shape[0]  # Observation Dimension

        self.t = 0  # Time Counter (for decaying learning rate)

        self.m, self.h = m, h

        self.lr_scale, self.decay = lr_scale, decay

        self.RM = RM

        # Construct truncated markov operator G
        self.G = jnp.zeros((h, d_obs, d_action))
        A_power = jnp.identity(d_state)
        for i in range(h):
            self.G = jax.ops.index_update(self.G, i, C @ A_power @ B)
            A_power = A_power @ A

        # Model Parameters
        # initial linear policy / perturbation contributions / bias
        self.K = K if K is not None else jnp.zeros((d_action, d_obs))

        self.M = lr_scale * jax.random.normal(
            self.random.generate_key(), shape=(m, d_action, d_obs)
        )

        # Past m nature y's such that y_nat[0] is the most recent
        self.y_nat = jnp.zeros((m, d_obs, 1))

        # Past h u's such that u[0] is the most recent
        self.us = jnp.zeros((h, d_action, 1))

        def policy_loss(M, G, y_nat, us):
            """Surrogate cost function"""

            def action(obs):
                """Action function"""
                return -self.K @ obs + jnp.tensordot(M, y_nat, axes=([0, 2], [0, 1]))

            final_state = y_nat[0] + jnp.tensordot(G, us, axes=([0, 2], [0, 1]))
            return cost_fn(final_state, action(final_state))

        self.policy_loss = policy_loss
        self.grad = jit(grad(policy_loss))

    def __call__(self, obs: jnp.ndarray) -> jnp.ndarray:
        """
        Description: Return the action based on current state and internal parameters.

        Args:
            state (jnp.ndarray): current state

        Returns:
           jnp.ndarray: action to take
        """
        # update y_nat
        self.update_noise(obs)

        # get action
        action = self.get_action(obs)

        # update Parameters
        self.update_params(obs, action)

        return action

    def get_action(self, obs: jnp.ndarray) -> jnp.ndarray:
        """
        Description: get action from state.

        Args:
            state (jnp.ndarray):

        Returns:
            jnp.ndarray
        """

        return -self.K @ obs + jnp.tensordot(self.M, self.y_nat, axes=([0, 2], [0, 1]))

    def update(self, obs: jnp.ndarray, u: jnp.ndarray) -> None:
        self.update_noise(obs)
        self.update_params(obs, u)

    def update_noise(self, obs: jnp.ndarray) -> None:
        y_nat = obs - jnp.tensordot(self.G, self.us, axes=([0, 2], [0, 1]))
        self.y_nat = jnp.roll(self.y_nat, 1, axis=0)
        self.y_nat = jax.ops.index_update(self.y_nat, 0, y_nat)

    def update_params(self, obs: jnp.ndarray, u: jnp.ndarray) -> None:
        """
        Description: update agent internal state.

        Args:
            state (jnp.ndarray):

        Returns:
            None
        """

        # update parameters
        delta_M = self.grad(self.M, self.G, self.y_nat, self.us)
        lr = self.lr_scale
        # lr *= (1/ (self.t+1)) if self.decay else 1
        lr = jax.lax.cond(self.decay, lambda x: x * 1 / (self.t + 1), lambda x: 1.0, lr)

        self.M -= lr * delta_M
        # if(jnp.linalg.norm(self.M) > self.RM):
        #     self.M *= (self.RM / jnp.linalg.norm(self.M))

        self.M = jax.lax.cond(
            jnp.linalg.norm(self.M) > self.RM,
            lambda x: x * (self.RM / jnp.linalg.norm(self.M)),
            lambda x: x,
            self.M,
        )

        # update us
        self.us = jnp.roll(self.us, 1, axis=0)
        self.us = jax.ops.index_update(self.us, 0, u)

        self.t += 1



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# TODO
# - interp smh
import jax.numpy as jnp

from deluca.core import JaxObject

DEFAULT_PRESSURE_RANGE = (5.0, 35.0)
DEFAULT_KEYPOINTS = [1e-8, 1.0, 1.5, 3.0]


class BreathWaveform(JaxObject):
    """Waveform generator with shape /‾\_"""

    def __init__(self, range=None, keypoints=None):
        self.lo, self.hi = range or DEFAULT_PRESSURE_RANGE
        self.xp = jnp.asarray([0] + (keypoints or DEFAULT_KEYPOINTS))
        self.fp = jnp.asarray([self.lo, self.hi, self.hi, self.lo, self.lo])
        self.period = self.xp[-1]

    def at(self, t):
        # return jnp.interp(t, self.xp, self.fp, period=self.period)
        return jnp.interp(t, self.xp, self.fp, period=3)

    def phase(self, t):
        return jnp.searchsorted(self.xp, t % self.period, side="right")


__all__ = ["BreathWaveform"]



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import jax.numpy as jnp

from deluca.core import Env
from deluca.core import field
from deluca.core import Obj


class CartpoleState(Obj):
    arr: jnp.ndarray = field(jaxed=True)
    h: int = field(0, jaxed=True)
    offset: float = field(0.0, jaxed=False)


class Cartpole(Env):
    m: float = field(0.1, jaxed=False)
    M: float = field(1.0, jaxed=False)
    l: float = field(1.0, jaxed=False)
    g: float = field(9.81, jaxed=False)
    dt: float = field(0.02, jaxed=False)
    H: int = field(10, jaxed=False)
    goal_state: jnp.ndarray = field(jnp.array([0.0, 0.0, 0.0, 0.0]), jaxed=False)
    dynamics: bool = field(False, jaxed=False)

    def init(self):
        return CartpoleState(arr=jnp.array([0.0, 0.0, 0.0, 0.0]))

    def __call__(self, state, action):
        A = jnp.array(
            [
                [1.0, 0.0, self.dt, 0.0],
                [0.0, 1.0, 0.0, self.dt],
                [0.0, self.dt * self.m * self.g / self.M, 1.0, 0.0],
                [0.0, self.dt * (self.m + self.M) * self.g / (self.M * self.l), 0.0, 1.0],
            ]
        )
        B = (jnp.array([[0.0], [0.0], [self.dt / self.M], [self.dt / (self.M * self.l)]]),)
        arr = (A @ state.arr + B @ (action + state.offset),)
        return state.replace(arr=arr, h=state.h + 1), arr



import jax
import jax.numpy as jnp

from deluca.agents.core import Agent
from deluca.utils.planning import f_at_x
from deluca.utils.planning import LQR
from deluca.utils.planning import rollout

# from deluca.agents._ilqr import LQR


def counterfact_loss(E, off, W, H, M, x, t, env_sim, U_old, k, K, X_old, D, F, w_is, alpha, C):
    y, cost = x, 0
    for h in range(H):
        u_delta = jnp.tensordot(E, W[h : h + M], axes=([0, 2], [0, 1])) + off
        u = U_old[t + h] + alpha * k[t + h] + K[t + h] @ (y - X_old[t + h]) + C * u_delta
        cost = env_sim.c(y, u)

        if w_is == "de":
            y = env_sim.f(y, u) + W[h + M]
        elif w_is == "dede":
            y = env_sim.f(y, u) + D[h + M] + W[h + M]
        else:
            y = (
                X_old[h + M + 1]
                + F[h + M][0] @ (y - X_old[h + M])
                + F[h + M][1] @ (u - U_old[h + M])
                + W[h + M]
            )
    return cost


grad_loss = jax.grad(counterfact_loss, (0, 1))


def rolls_by_igpc(env_true, env_sim, U_old, k, K, X_old, D, F, alpha, w_is, ref_lr=0.1):
    # print(ref_lr)
    H, M, lr, C = 3, 3, ref_lr, 1.0
    X, U, U_delta = (
        [None for _ in range(env_sim.H + 1)],
        [None for _ in range(env_sim.H)],
        [None for _ in range(env_sim.H)],
    )
    W, E, off = (
        jnp.zeros((H + M, env_sim.state_dim)),
        jnp.zeros((H, env_sim.action_dim, env_sim.state_dim)),
        jnp.zeros(env_sim.action_dim),
    )
    X[0], cost = env_true.reset(), 0
    for h in range(env_true.H):
        U_delta[h] = jnp.tensordot(E, W[-M:], axes=([0, 2], [0, 1])) + off
        U[h] = U_old[h] + alpha * k[h] + K[h] @ (X[h] - X_old[h]) + C * U_delta[h]
        X[h + 1], instant_cost, _, _ = env_true.step(U[h])
        cost += instant_cost

        # 3 choices for w are f-g, (f-g)-(f-g last time), (f- LIN(g) with OFF).
        if w_is == "de":
            w = X[h + 1] - env_sim.f(X[h], U[h])
        elif w_is == "dede":
            w = (X[h + 1] - env_sim.f(X[h], U[h])) - D[h]
        else:
            w = X[h + 1] - (
                X_old[h + 1] + F[h][0] @ (X[h] - X_old[h]) + F[h][1] @ (U[h] - U_old[h])
            )

        W = jax.ops.index_update(W, 0, w)
        W = jnp.roll(W, -1, axis=0)
        if h >= H:
            delta_E, delta_off = grad_loss(
                E, off, W, H, M, X[h - H], h - H, env_sim, U_old, k, K, X_old, D, F, w_is, alpha, C,
            )
            E -= lr / h * delta_E
            off -= lr / h * delta_off
    # print(norm_U_old, norm_other, norm_off, norm_U_delta, norm_U_vals)
    return X, U, cost


def igpc_closed(
    env_true, env_sim, U, T, k=None, K=None, X=None, w_is="de", lr=None, ref_alpha=1.0, log=None
):
    alpha, r = ref_alpha, 1
    X, U, c = rollout(env_true, U, k, K, X)
    print("initial cost:" + str(c))
    assert w_is in ["de", "dede", "delin"]
    for t in range(T):
        D, F, C = f_at_x(env_sim, X, U)
        k, K = LQR(F, C)
        for alphaC in alpha * 1.1 * 1.1 ** (-jnp.arange(10) ** 2):
            r += 1
            XC, UC, cC = rolls_by_igpc(env_true, env_sim, U, k, K, X, D, F, alphaC, w_is, lr)
            if cC <= c:
                X, U, c, alpha = XC, UC, cC, alphaC
                print(f"(iGPC): t = {t}, r = {r}, c = {c}, alpha = {alpha}")
                if log is not None:
                    log.append(f"(iGPC): t = {t}, r = {r}, c = {c}")
                break
        else:
            return X, U, k, K, c
    return X, U, k, K, c


class IGPCInner:
    def __init__(self, env_true, env_sim):
        self.env_true = env_true
        self.env_sim = env_sim

    def train(self, T, U_init=None, k=None, K=None, X=None, w_is="de", lr=None, ref_alpha=1.0):
        self.log = []
        if U_init is None:
            U_init = [jnp.zeros(self.env_true.action_dim) for _ in range(self.env_true.H)]
        X, U, k, K, c = igpc_closed(
            self.env_true, self.env_sim, U_init, T, k, K, X, w_is, lr, ref_alpha, self.log
        )
        return X, U, k, K, c, self.log


class IGPC(Agent):
    def __init__(self):
        self.reset()

    def train(
        self,
        env_true,
        env_sim,
        train_steps,
        U_init=None,
        k=None,
        K=None,
        X=None,
        w_is="de",
        lr=None,
        ref_alpha=1.0,
    ):
        self.env_true = env_true
        self.env_sim = env_sim
        igpc_agent = IGPCInner(self.env_true, self.env_sim)
        X, U, k, K, c, log = igpc_agent.train(train_steps, U_init, k, K, X, w_is, lr, ref_alpha)
        self.U = jnp.array(U)
        self.X = jnp.array(X)
        self.k = jnp.array(k)
        self.K = jnp.array(K)
        return c, log

    def reset(self):
        self.t = 0

    def __call__(self, state):
        action = self.U[self.t]
        self.t += 1

        return action



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import jax.numpy as jnp

from deluca.core import Env
from deluca.core import field
from deluca.core import Obj


class ReacherState(Obj):
    arr: jnp.ndarray = field(jaxed=True)
    h: float = field(0.0, jaxed=True)

    def flatten(self):
        return self.arr

    def unflatten(self, arr):
        return ReacherState(arr=arr, h=self.h)


class Reacher(Env):
    m1: float = field(1.0, jaxed=False)
    m2: float = field(1.0, jaxed=False)
    l1: float = field(1.0, jaxed=False)
    l2: float = field(1.0, jaxed=False)
    g: float = field(0.0, jaxed=False)
    max_torque: float = field(1.0, jaxed=False)
    dt: float = field(0.01, jaxed=False)
    H: int = field(200, jaxed=False)
    goal_coord: jnp.ndarray = field(jnp.array([0.0, 1.8]), jaxed=False)
    state_dim: float = field(6, jaxed=False)
    action_dim: float = field(2, jaxed=False)

    def init(self):
        initial_th = (jnp.pi / 4, jnp.pi / 2)
        return ReacherState(
            arr=jnp.array(
                [
                    *initial_th,
                    0.0,
                    0.0,
                    self.l1 * jnp.cos(initial_th[0])
                    + self.l2 * jnp.cos(initial_th[0] + initial_th[1])
                    - self.goal_coord[0],
                    self.l1 * jnp.sin(initial_th[0])
                    + self.l2 * jnp.sin(initial_th[0] + initial_th[1])
                    - self.goal_coord[1],
                ]
            )
        )

    def __call__(self, state, action):
        m1, m2, l1, l2, g = self.m1, self.m2, self.l1, self.l2, self.g
        th1, th2, dth1, dth2, Dx, Dy = state.arr
        t1, t2 = action

        a11 = (m1 + m2) * l1 ** 2 + m2 * l2 ** 2 + 2 * m2 * l1 * l2 * jnp.cos(th2)
        a12 = m2 * l2 ** 2 + m2 * l1 * l2 * jnp.cos(th2)
        a22 = m2 * l2 ** 2
        b1 = (
            t1
            + m2 * l1 * l2 * (2 * dth1 + dth2) * dth2 * jnp.sin(th2)
            - m2 * l2 * g * jnp.sin(th1 + th2)
            - (m1 + m2) * l1 * g * jnp.sin(th1)
        )
        b2 = t2 - m2 * l1 * l2 * dth1 ** 2 * jnp.sin(th2) - m2 * l2 * g * jnp.sin(th1 + th2)
        A, b = jnp.array([[a11, a12], [a12, a22]]), jnp.array([b1, b2])
        ddth1, ddth2 = jnp.linalg.inv(A) @ b

        th1, th2 = th1 + dth1 * self.dt, th2 + dth2 * self.dt
        dth1, dth2 = dth1 + ddth1 * self.dt, dth2 + ddth2 * self.dt
        Dx, Dy = (
            l1 * jnp.cos(th1) + l2 * jnp.cos(th1 + th2) - self.goal_coord[0],
            l1 * jnp.sin(th1) + l2 * jnp.sin(th1 + th2) - self.goal_coord[1],
        )

        arr = jnp.array([th1, th2, dth1, dth2, Dx, Dy])
        new_state = state.replace(arr=arr, h=state.h + 1)

        return new_state, arr



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Callable

import jax

from deluca.core import Agent
from deluca.core import field


class Random(Agent):
    seed: int = field(0, jaxed=False)
    func: Callable = field(lambda key: 0.0, jaxed=False)

    def init(self):
        return jax.random.PRNGKey(self.seed)

    def __call__(self, state, obs):
        key, subkey = jax.random.split(state)

        return subkey, self.func(key)



import os

import jax.numpy as np
from PIL import Image


def f_at_x(env, X, U, H=None):
    if H is None:
        H = env.H
    D, F, C = (
        [None for _ in range(H)],
        [None for _ in range(H)],
        [None for _ in range(H)],
    )
    for h in range(H):
        D[h] = X[h + 1] - env.f(X[h], U[h])
        F[h] = env.f_x(X[h], U[h]), env.f_u(X[h], U[h])
        C[h] = (
            env.c_x(X[h], U[h]),
            env.c_u(X[h], U[h]),
            env.c_xx(X[h], U[h]),
            env.c_uu(X[h], U[h]),
        )
    return D, F, C


def rollout(
    env,
    U_old,
    k=None,
    K=None,
    X_old=None,
    alpha=1.0,
    D=None,
    F=None,
    H=None,
    start_state=None,
    render=False,
    render_dir=".",
):
    """
    Arg List
    env: The environment to do the rollout on. This is treated as a derstructible copy.
    U_old: A base open loop control sequence
    k: open loop gain (iLQR)
    K: closed loop gain (iLQR)
    X_old: Previous trajectory to compute gain (iLQR)
    alpha: Optional multplier to the open loop gain (iLQR)
    D: Optional noise vectors for the rollout (GPC)
    F: Linearization shift ??? (GPC)
    H: The horizon length to perform a rollout.
    render: Whether to render the rollout or not
    render_dir: Which directory to render the rollout.
    """
    if render and not os.path.exists(render_dir):
        os.makedirs(render_dir)
    if H is None:
        H = env.H
    X, U = [None for _ in range(H + 1)], [None for _ in range(H)]
    if start_state is not None:
        X[0], cost = env.reset_model(start_state), 0.0
    else:
        X[0], cost = env.reset(), 0.0
    for h in range(H):
        if k is None:
            U[h] = U_old[h]
        else:
            U[h] = U_old[h] + alpha * k[h] + K[h] @ (X[h] - X_old[h])

        if D is None:
            X[h + 1], instant_cost, _, _ = env.step(U[h])
        elif F is None:
            X[h + 1], instant_cost, _, _ = env.step(U[h])
            X[h + 1] += D[h]
        else:
            X[h + 1] = X_old[h + 1] + F[h][0] @ (X[h] - X_old[h]) + F[h][1] @ (U[h] - U_old[h])
            instant_cost = env.c(X[h], U[h], h)

        cost += instant_cost
        if render:
            img = env.render(mode="rgb_array")
            img = Image.fromarray(img)
            img.save("{}/{}.png".format(render_dir, h), "PNG")
    return X, U, cost


def LQR(F, C):
    H, d = len(C), C[0][0].shape[0]
    K, k = (
        [None for _ in range(H)],
        [None for _ in range(H)],
    )
    V_x, V_xx = np.zeros(d), np.zeros((d, d))
    for h in range(H - 1, -1, -1):
        f_x, f_u = F[h]
        c_x, c_u, c_xx, c_uu = C[h]
        Q_x, Q_u = c_x + f_x.T @ V_x, c_u + f_u.T @ V_x
        Q_xx, Q_ux, Q_uu = (
            c_xx + f_x.T @ V_xx @ f_x,
            f_u.T @ V_xx @ f_x,
            c_uu + f_u.T @ V_xx @ f_u,
        )

        K[h], k[h] = -np.linalg.inv(Q_uu) @ Q_ux, -np.linalg.inv(Q_uu) @ Q_u
        V_x = Q_x - K[h].T @ Q_uu @ k[h]
        V_xx = Q_xx - K[h].T @ Q_uu @ K[h]
        V_xx = (V_xx + V_xx.T) / 2
    return k, K



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""deluca.agents._bpc"""
from numbers import Real

import jax
import jax.numpy as jnp
import numpy as np
import numpy.random as random

from deluca.agents._lqr import LQR
from deluca.agents.core import Agent


def generate_uniform(shape, norm=1.00):
    v = random.normal(size=shape)
    v = norm * v / np.linalg.norm(v)
    v = np.array(v)
    return v


class BPC(Agent):
    def __init__(
        self,
        A: jnp.ndarray,
        B: jnp.ndarray,
        Q: jnp.ndarray = None,
        R: jnp.ndarray = None,
        K: jnp.ndarray = None,
        start_time: int = 0,
        H: int = 5,
        lr_scale: Real = 0.005,
        decay: bool = False,
        delta: Real = 0.01,
    ) -> None:
        """
        Description: Initialize the dynamics of the model.

        Args:
            A (jnp.ndarray): system dynamics
            B (jnp.ndarray): system dynamics
            Q (jnp.ndarray): cost matrices (i.e. cost = x^TQx + u^TRu)
            R (jnp.ndarray): cost matrices (i.e. cost = x^TQx + u^TRu)
            K (jnp.ndarray): Starting policy (optional). Defaults to LQR gain.
            start_time (int):
            H (postive int): history of the controller
            lr_scale (Real):
            decay (boolean):
        """

        self.d_state, self.d_action = B.shape  # State & Action Dimensions

        self.A, self.B = A, B  # System Dynamics

        self.t = 0  # Time Counter (for decaying learning rate)

        self.H = H

        self.lr_scale, self.decay = lr_scale, decay

        self.delta = delta

        # Model Parameters
        # initial linear policy / perturbation contributions / bias
        # TODO: need to address problem of LQR with jax.lax.scan
        self.K = K if K is not None else LQR(self.A, self.B, Q, R).K

        self.M = self.delta * generate_uniform((H, self.d_action, self.d_state))

        # Past H noises ordered increasing in time
        self.noise_history = jnp.zeros((H, self.d_state, 1))

        # past state and past action
        self.state, self.action = jnp.zeros((self.d_state, 1)), jnp.zeros((self.d_action, 1))

        self.eps = generate_uniform((H, H, self.d_action, self.d_state))
        self.eps_bias = generate_uniform((H, self.d_action, 1))

        def grad(M, noise_history, cost):
            return cost * jnp.sum(self.eps, axis=0)

        self.grad = grad

    def __call__(self, state: jnp.ndarray, cost: Real) -> jnp.ndarray:
        """
        Description: Return the action based on current state and internal parameters.

        Args:
            state (jnp.ndarray): current state

        Returns:
           jnp.ndarray: action to take
        """

        action = self.get_action(state)
        self.update(state, action, cost)
        return action

    def update(self, state: jnp.ndarray, action: jnp.ndarray, cost: Real) -> None:
        """
        Description: update agent internal state.

        Args:
            state (jnp.ndarray): current state
            action (jnp.ndarray): action taken
            cost (Real): scalar cost received

        Returns:
            None
        """
        noise = state - self.A @ self.state - self.B @ action
        self.noise_history = jax.ops.index_update(self.noise_history, 0, noise)
        self.noise_history = jnp.roll(self.noise_history, -1, axis=0)

        lr = self.lr_scale
        lr *= (1 / (self.t ** (3 / 4) + 1)) if self.decay else 1

        delta_M = self.grad(self.M, self.noise_history, cost)
        self.M -= lr * delta_M

        self.eps = jax.ops.index_update(
            self.eps, 0, generate_uniform((self.H, self.d_action, self.d_state))
        )
        self.eps = np.roll(self.eps, -1, axis=0)

        self.M += self.delta * self.eps[-1]

        # update state
        self.state = state

        self.t += 1

    def get_action(self, state: jnp.ndarray) -> jnp.ndarray:
        """
        Description: get action from state.

        Args:
            state (jnp.ndarray):

        Returns:
            jnp.ndarray
        """
        return -self.K @ state + jnp.tensordot(self.M, self.noise_history, axes=([0, 2], [0, 1]))



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import brax
from brax.io import html
from IPython.display import HTML

from deluca.core import Env
from deluca.core import field


class BraxEnv(Env):
    sys: brax.System = field(jaxed=False)

    def setup(self):
        if isinstance(self.sys, brax.physics.config_pb2.Config):
            self.sys = brax.System(self.sys)

    def init(self):
        return self.sys.default_qp()

    def __call__(self, state, action):
        state, _ = self.sys.step(state, action)
        return state, state

    def render(self, states):
        return HTML(html.render(self.sys, states))



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import jax
import jax.numpy as jnp

from deluca.core import Env
from deluca.core import field

__copyright__ = "Copyright 2013, RLPy http://acl.mit.edu/RLPy"
__credits__ = [
    "Alborz Geramifard",
    "Robert H. Klein",
    "Christoph Dann",
    "William Dabney",
    "Jonathan P. How",
]
__license__ = "BSD 3-Clause"
__author__ = "Christoph Dann <cdann@cdann.de>"

# SOURCE:
# https://github.com/rlpy/rlpy/blob/master/rlpy/Domains/Acrobot.py


class Acrobot(Env):
    key: jnp.ndarray = field(jax.random.PRNGKey(0), jaxed=False)
    dt: float = field(0.2, jaxed=False)

    LINK_LENGTH_1: float = field(1.0, jaxed=False)
    LINK_LENGTH_2: float = field(1.0, jaxed=False)
    LINK_MASS_1: float = field(1.0, jaxed=False)
    LINK_MASS_2: float = field(1.0, jaxed=False)
    LINK_COM_POS_1: float = field(0.0, jaxed=False)
    LINK_COM_POS_2: float = field(0.0, jaxed=False)
    LINK_MOI: float = field(1.0)

    MAX_VEL_1: float = field(4 * jnp.pi, jaxed=False)
    MAX_VEL_2: float = field(9 * jnp.pi, jaxed=False)

    AVAIL_TORQUE: jnp.ndarray = field(jnp.array([-1.0, 0.0, +1]), jaxed=False)

    torque_noise_max: float = field(0.0, jaxed=False)

    high: jnp.ndarray = field(jaxed=False)
    low: jnp.ndarray = field(jaxed=False)

    def setup(self):
        self.high = jnp.array([1.0, 1.0, 1.0, 1.0, self.MAX_VEL_1, self.MAX_VEL_2])
        self.low = -self.high

    def __call__(self, state, action):
        augmented_state = jnp.append(state, action)

        new_state = rk4(self._dsdt, augmented_state, [0, self.dt])
        # only care about final timestep of integration returned by integrator
        new_state = new_state[-1]
        new_state = new_state[:4]  # omit action
        # ODEINT IS TOO SLOW!
        # ns_continuous = integrate.odeint(self._dsdt, self.s_continuous, [0, self.dt])
        # self.s_continuous = ns_continuous[-1] # We only care about the state
        # at the ''final timestep'', self.dt

        new_state = jax.ops.index_update(new_state, 0, wrap(new_state[0], -jnp.pi, jnp.pi))
        new_state = jax.ops.index_update(new_state, 1, wrap(new_state[1], -jnp.pi, jnp.pi))
        new_state = jax.ops.index_update(
            new_state, 2, bound(new_state[2], -self.MAX_VEL_1, self.MAX_VEL_1)
        )
        new_state = jax.ops.index_update(
            new_state, 3, bound(new_state[3], -self.MAX_VEL_2, self.MAX_VEL_2)
        )

        return (
            new_state,
            jnp.array(
                [
                    jnp.cos(new_state[0]),
                    jnp.sin(new_state[0]),
                    jnp.cos(new_state[1]),
                    jnp.sin(new_state[1]),
                    new_state[2],
                    new_state[3],
                ]
            ),
        )

    def _dsdt(self, augmented_state, t):
        m1 = self.LINK_MASS_1
        m2 = self.LINK_MASS_2
        l1 = self.LINK_LENGTH_1
        lc1 = self.LINK_COM_POS_1
        lc2 = self.LINK_COM_POS_2
        I1 = self.LINK_MOI
        I2 = self.LINK_MOI
        g = 9.8
        a = augmented_state[-1]
        s = augmented_state[:-1]

        theta1 = s[0]
        theta2 = s[1]
        dtheta1 = s[2]
        dtheta2 = s[3]

        d1 = m1 * lc1 ** 2 + m2 * (l1 ** 2 + lc2 ** 2 + 2 * l1 * lc2 * jnp.cos(theta2)) + I1 + I2
        d2 = m2 * (lc2 ** 2 + l1 * lc2 * jnp.cos(theta2)) + I2
        phi2 = m2 * lc2 * g * jnp.cos(theta1 + theta2 - jnp.pi / 2.0)
        phi1 = (
            -m2 * l1 * lc2 * dtheta2 ** 2 * jnp.sin(theta2)
            - 2 * m2 * l1 * lc2 * dtheta2 * dtheta1 * jnp.sin(theta2)
            + (m1 * lc1 + m2 * l1) * g * jnp.cos(theta1 - jnp.pi / 2)
            + phi2
        )
        ddtheta2 = (a + d2 / d1 * phi1 - phi2) / (m2 * lc2 ** 2 + I2 - d2 ** 2 / d1)
        # if self.book_or_nips == "nips":
        # the following line is consistent with the description in the
        # paper
        # ddtheta2 = (a + d2 / d1 * phi1 - phi2) / (m2 * lc2 ** 2 + I2 - d2 ** 2 / d1)
        # else:
        # the following line is consistent with the java implementation and the
        # book
        # ddtheta2 = (
        # a + d2 / d1 * phi1 - m2 * l1 * lc2 * dtheta1 ** 2 * jnp.sin(theta2) - phi2
        # ) / (m2 * lc2 ** 2 + I2 - d2 ** 2 / d1)
        ddtheta1 = -(d2 * ddtheta2 + phi1) / d1
        return (dtheta1, dtheta2, ddtheta1, ddtheta2, 0.0)  # 4-state version


def wrap(x, m, M):
    """Wraps ``x`` so m <= x <= M; but unlike ``bound()`` which
    truncates, ``wrap()`` wraps x around the coordinate system defined by m,M.\n
    For example, m = -180, M = 180 (degrees), x = 360 --> returns 0.
    Args:
        x: a scalar
        m: minimum possible value in range
        M: maximum possible value in range
    Returns:
        x: a scalar, wrapped
    """
    diff = M - m

    if diff == 0:
        return x

    to_subtract = jnp.ceil(jnp.maximum(0, x - M) / diff)
    x -= to_subtract * diff

    to_add = jnp.ceil(jnp.maximum(0, m - x) / diff)
    x += to_add * diff

    return x


def bound(x, m, M=None):
    """Either have m as scalar, so bound(x,m,M) which returns m <= x <= M *OR*
    have m as length 2 vector, bound(x,m, <IGNORED>) returns m[0] <= x <= m[1].
    Args:
        x: scalar
    Returns:
        x: scalar, bound between min (m) and Max (M)
    """
    if M is None:
        M = m[1]
        m = m[0]
    # bound x between min (m) and Max (M)
    return jnp.minimum(jnp.maximum(x, m), M)


def rk4(derivs, y0, t, *args, **kwargs):
    """
    Integrate 1D or ND system of ODEs using 4-th order Runge-Kutta.
    This is a toy implementation which may be useful if you find
    yourself stranded on a system w/o scipy.  Otherwise use
    :func:`scipy.integrate`.
    Args:
        derivs: the derivative of the system and has the signature ``dy = derivs(yi, ti)``
        y0: initial state vector
        t: sample times
        args: additional arguments passed to the derivative function
        kwargs: additional keyword arguments passed to the derivative function
    Example 1 ::
        ## 2D system
        def derivs6(x,t):
            d1 =  x[0] + 2*x[1]
            d2 =  -3*x[0] + 4*x[1]
            return (d1, d2)
        dt = 0.0005
        t = arange(0.0, 2.0, dt)
        y0 = (1,2)
        yout = rk4(derivs6, y0, t)
    Example 2::
        ## 1D system
        alpha = 2
        def derivs(x,t):
            return -alpha*x + exp(-t)
        y0 = 1
        yout = rk4(derivs, y0, t)
    If you have access to scipy, you should probably be using jnp.sing the
    scipy.integrate tools rather than this function.
    Returns:
        yout: Runge-Kutta approximation of the ODE
    """

    try:
        Ny = len(y0)
    except TypeError:
        yout = jnp.zeros((len(t),))
    else:
        yout = jnp.zeros((len(t), Ny))

    yout = jax.ops.index_update(yout, 0, y0)

    for i in jnp.arange(len(t) - 1):

        thist = t[i]
        dt = t[i + 1] - thist
        dt2 = dt / 2.0
        y0 = yout[i]

        k1 = jnp.asarray(derivs(y0, thist, *args, **kwargs))
        k2 = jnp.asarray(derivs(y0 + dt2 * k1, thist + dt2, *args, **kwargs))
        k3 = jnp.asarray(derivs(y0 + dt2 * k2, thist + dt2, *args, **kwargs))
        k4 = jnp.asarray(derivs(y0 + dt * k3, thist + dt, *args, **kwargs))
        yout = jax.ops.index_update(yout, i + 1, y0 + dt / 6.0 * (k1 + 2 * k2 + 2 * k3 + k4))

    return yout



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from deluca.envs._brax import BraxEnv
from deluca.envs.classic._acrobot import Acrobot
from deluca.envs.classic._cartpole import Cartpole
from deluca.envs.classic._mountain_car import MountainCar
from deluca.envs.classic._pendulum import Pendulum
from deluca.envs.classic._planar_quadrotor import PlanarQuadrotor
from deluca.envs.classic._reacher import Reacher

__all__ = [
    "BraxEnv",
    "Acrobot",
    "Cartpole",
    "MountainCar",
    "Pendulum",
    "PlanarQuadrotor",
    "Reacher",
]



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License
"""
@author: Olivier Sigaud
A merge between two sources:
* Adaptation of the MountainCar Environment from the "FAReinforcement" library
of Jose Antonio Martin H. (version 1.0), adapted by  'Tom Schaul, tom@idsia.ch'
and then modified by Arnaud de Broissia
* the OpenAI/gym MountainCar environment
itself from
http://incompleteideas.net/sutton/MountainCar/MountainCar1.cp
permalink: https://perma.cc/6Z2N-PFWC
"""
import jax
import jax.numpy as jnp

from deluca.core import Env
from deluca.core import field


class MountainCar(Env):
    key: jnp.ndarray = field(jax.random.PRNGKey(0), jaxed=False)
    goal_velocity: float = field(0.0, jaxed=False)
    min_action: float = field(-1.0, jaxed=False)
    max_action: float = field(1.0, jaxed=False)
    min_position: float = field(-1.2, jaxed=False)
    max_position: float = field(0.6, jaxed=False)
    max_speed: float = field(0.07, jaxed=False)
    goal_position: float = field(0.5, jaxed=False)
    power = 0.0015

    low_state: jnp.ndarray = field(jaxed=False)
    high_state: jnp.ndarray = field(jaxed=False)

    def setup(self):
        self.low_state = jnp.array([self.min_position, -self.max_speed])
        self.high_state = jnp.array([self.max_position, self.max_speed])

    def init(self):
        return jnp.array([jax.random.uniform(self.key, min_val=-0.6, maxval=0.4), 0])

    def __call__(self, state, action):
        position, velocity = state

        force = jnp.minimum(jnp.maximum(action, self.min_action), self.max_action)

        velocity += force * self.power - 0.0025 * jnp.cos(3 * position)
        velocity = jnp.clip(velocity, -self.max_speed, self.max_speed)

        position += velocity
        position = jnp.clip(position, self.min_position, self.max_position)
        reset_velocity = (position == self.min_position) & (velocity < 0)
        velocity = jax.lax.cond(
            reset_velocity[0], velocity, lambda x: jnp.zeros((1,)), velocity, lambda x: x
        )
        new_state = jnp.reshape(jnp.array([position, velocity]), (2,))

        return new_state, new_state



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""deluca.agents._adaptive"""
from numbers import Real
from typing import Callable

import jax
import jax.numpy as jnp
import numpy as np

from deluca.agents.core import Agent


def quad_loss(x: jnp.ndarray, u: jnp.ndarray) -> Real:
    """
    Quadratic loss.

    Args:
        x (jnp.ndarray):
        u (jnp.ndarray):

    Returns:
        Real
    """
    return jnp.sum(x.T @ x + u.T @ u)


def lifetime(x, lower_bound):
    l = 4
    while x % 2 == 0:
        l *= 2
        x /= 2

    return max(lower_bound, l + 1)


class Adaptive(Agent):
    def __init__(
        self,
        T: int,
        base_controller,
        A: jnp.ndarray,
        B: jnp.ndarray,
        cost_fn: Callable[[jnp.ndarray, jnp.ndarray], Real] = None,
        HH: int = 10,
        eta: Real = 0.5,
        eps: Real = 1e-6,
        inf: Real = 1e6,
        life_lower_bound: int = 100,
        expert_density: int = 64,
    ) -> None:

        self.A, self.B = A, B
        self.n, self.m = B.shape

        cost_fn = cost_fn or quad_loss

        self.base_controller = base_controller

        # Start From Uniform Distribution
        self.T = T
        self.weights = np.zeros(T)
        self.weights[0] = 1.0

        # Track current timestep
        self.t, self.expert_density = 0, expert_density

        # Store Model Hyperparameters
        self.eta, self.eps, self.inf = eta, eps, inf

        # State and Action
        self.x, self.u = jnp.zeros((self.n, 1)), jnp.zeros((self.m, 1))

        # Alive set
        self.alive = jnp.zeros((T,))

        # Precompute time of death at initialization
        self.tod = np.arange(T)
        for i in range(1, T):
            self.tod[i] = i + lifetime(i, life_lower_bound)
        self.tod[0] = life_lower_bound  # lifetime not defined for 0

        # Maintain Dictionary of Active Learners
        self.learners = {}
        self.learners[0] = base_controller(A, B, cost_fn=cost_fn)

        self.w = jnp.zeros((HH, self.n, 1))

        def policy_loss(controller, A, B, x, w):
            def evolve(x, h):
                """Evolve function"""
                return A @ x + B @ controller.get_action(x) + w[h], None

            final_state, _ = jax.lax.scan(evolve, x, jnp.arange(HH))
            return cost_fn(final_state, controller.get_action(final_state))

        self.policy_loss = policy_loss

    def __call__(self, x, A, B):

        play_i = np.argmax(self.weights)
        self.u = self.learners[play_i].get_action(x)

        # Update alive models
        for i in jnp.nonzero(self.alive)[0]:
            loss_i = self.policy_loss(self.learners[i], A, B, x, self.w)
            self.weights[i] *= np.exp(-self.eta * loss_i)
            self.weights[i] = min(max(self.weights[i], self.eps), self.inf)
            self.learners[i].update(x, u=self.u)

        self.t += 1

        # One is born every expert_density steps
        if self.t % self.expert_density == 0:
            self.alive = jax.ops.index_update(self.alive, self.t, 1)
            self.weights[self.t] = self.eps
            self.learners[self.t] = self.base_controller(A, B, cost_fn=self.cost_fn)
            self.learners[self.t].x = x

        # At most one dies
        kill_list = jnp.where(self.tod == self.t)
        if len(kill_list[0]):
            kill = int(kill_list[0][0])
            if self.alive[kill]:
                self.alive = jax.ops.index_update(self.alive, kill, 0)
                del self.learners[kill]
                self.weights[kill] = 0

        # Rescale
        max_w = np.max(self.weights)
        if max_w < 1:
            self.weights /= max_w

        # Get new noise (will be located at w[-1])
        self.w = jax.ops.index_update(self.w, 0, x - self.A @ self.x + self.B @ self.u)
        self.w = jnp.roll(self.w, -1, axis=0)

        # Update System
        self.x, self.A, self.B = x, A, B

        return self.u



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""deluca.agents._deep"""
from numbers import Real

import jax
import jax.numpy as jnp

from deluca.agents.core import Agent
from deluca.utils import Random


class Deep(Agent):
    """
    Generic deep controller that uses zero-order methods to train on an
    environment.
    """

    def __init__(
        self,
        env_state_size,
        action_space,
        learning_rate: Real = 0.001,
        gamma: Real = 0.99,
        max_episode_length: int = 500,
        seed: int = 0,
    ) -> None:
        """
        Description: initializes the Deep agent

        Args:
            env (Env): a deluca environment
            learning_rate (Real):
            gamma (Real):
            max_episode_length (int):
            seed (int):

        Returns:
            None
        """
        # Create gym and seed numpy
        self.env_state_size = int(env_state_size)
        self.action_space = action_space
        self.max_episode_length = max_episode_length
        self.lr = learning_rate
        self.gamma = gamma

        self.random = Random(seed)
        self.reset()

    def reset(self) -> None:
        """
        Description: reset agent

        Args:
            None

        Returns:
            None
        """
        # Init weight
        self.W = jax.random.uniform(
            self.random.generate_key(),
            shape=(self.env_state_size, len(self.action_space)),
            minval=0,
            maxval=1,
        )

        # Keep stats for final print of graph
        self.episode_rewards = []

        self.current_episode_length = 0
        self.current_episode_reward = 0
        self.episode_rewards = jnp.zeros(self.max_episode_length)
        self.episode_grads = jnp.zeros((self.max_episode_length, self.W.shape[0], self.W.shape[1]))

        # dummy values for attrs, needed to inform scan of traced shapes
        self.state = jnp.zeros((self.env_state_size,))
        self.action = self.action_space[0]
        ones = jnp.ones((len(self.action_space),))
        self.probs = ones * 1 / jnp.sum(ones)

    def policy(self, state: jnp.ndarray, w: jnp.ndarray) -> jnp.ndarray:
        """
        Description: Policy that maps state to action parameterized by w

        Args:
            state (jnp.ndarray):
            w (jnp.ndarray):
        """
        z = jnp.dot(state, w)
        exp = jnp.exp(z)
        return exp / jnp.sum(exp)

    def softmax_grad(self, softmax: jnp.ndarray) -> jnp.ndarray:
        """
        Description: Vectorized softmax Jacobian

        Args:
            softmax (jnp.ndarray)
        """
        s = softmax.reshape(-1, 1)
        return jnp.diagflat(s) - jnp.dot(s, s.T)

    def __call__(self, state: jnp.ndarray):
        """
        Description: provide an action given a state

        Args:
            state (jnp.ndarray):

        Returns:
            jnp.ndarray: action to take
        """
        self.state = state
        self.probs = self.policy(state, self.W)
        self.action = jax.random.choice(
            self.random.generate_key(), a=self.action_space, p=self.probs
        )

        return self.action

    def feed(self, reward: Real) -> None:
        """
        Description: compute gradient and save with reward in memory for weight updates

        Args:
            reward (Real):

        Returns:
            None
        """
        dsoftmax = self.softmax_grad(self.probs)[self.action, :]
        dlog = dsoftmax / self.probs[self.action]
        grad = self.state.reshape(-1, 1) @ dlog.reshape(1, -1)

        self.episode_rewards = jax.ops.index_update(
            self.episode_rewards, self.current_episode_length, reward
        )
        self.episode_grads = jax.ops.index_update(
            self.episode_grads, self.current_episode_length, grad
        )
        self.current_episode_length += 1

    def update(self) -> None:
        """
        Description: update weights
        
        Args:
            None

        Returns:
            None
        """
        for i in range(self.current_episode_length):
            # Loop through everything that happend in the episode and update
            # towards the log policy gradient times **FUTURE** reward
            self.W += self.lr * self.episode_grads[i] + jnp.sum(
                jnp.array(
                    [
                        r * (self.gamma ** r)
                        for r in self.episode_rewards[i : self.current_episode_length]
                    ]
                )
            )

        # reset episode length
        self.current_episode_length = 0



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import jax
import jax.numpy as jnp

from deluca.envs.lung import BreathWaveform
from deluca.envs.lung.core import Lung


def PropValve(x):
    y = 3.0 * x
    flow_new = 1.0 * (jnp.tanh(0.03 * (y - 130)) + 1.0)
    flow_new = jnp.clip(flow_new, 0.0, 1.72)
    return flow_new


def Solenoid(x):
    return x > 0


class BalloonLung(Lung):
    """
    Lung simulator based on the two-balloon experiment

    Source: https://en.wikipedia.org/wiki/Two-balloon_experiment

    TODO:
    - time / dt
    - dynamics
    - waveform
    - phase
    """

    def __init__(
        self,
        leak=False,
        peep_valve=5.0,
        PC=40.0,
        P0=0.0,
        C=10.0,
        R=15.0,
        dt=0.03,
        waveform=None,
        reward_fn=None,
    ):
        self.viewer = None
        # dynamics hyperparameters
        self.min_volume = 1.5
        self.C = C
        self.R = R
        self.PC = PC
        self.P0 = 0.0
        self.leak = leak
        self.peep_valve = peep_valve
        self.time = 0.0
        self.dt = dt

        self.waveform = waveform or BreathWaveform()

        self.r0 = (3.0 * self.min_volume / (4.0 * jnp.pi)) ** (1.0 / 3.0)

        # reset states
        self.reset()

    def reset(self):
        self.time = 0.0
        self.target = self.waveform.at(self.time)
        self.state = self.dynamics({"volume": self.min_volume, "pressure": -1.0}, (0.0, 0.0))
        return self.observation

    @property
    def observation(self):
        return {
            "measured": self.state["pressure"],
            "target": self.target,
            "dt": self.dt,
            "phase": self.waveform.phase(self.time),
        }

    def dynamics(self, state, action):
        """
        state: (volume, pressure)
        action: (u_in, u_out)
        """
        volume, pressure = state["volume"], state["pressure"]
        u_in, u_out = action

        flow = jnp.clip(PropValve(u_in) * self.R, 0.0, 2.0)
        flow -= jax.lax.cond(
            pressure > self.peep_valve,
            lambda x: jnp.clip(Solenoid(u_out), 0.0, 2.0) * 0.05 * pressure,
            lambda x: 0.0,
            flow,
        )

        volume += flow * self.dt
        volume += jax.lax.cond(
            self.leak,
            lambda x: (self.dt / (5.0 + self.dt) * (self.min_volume - volume)),
            lambda x: 0.0,
            0.0,
        )

        r = (3.0 * volume / (4.0 * jnp.pi)) ** (1.0 / 3.0)
        pressure = self.P0 + self.PC * (1.0 - (self.r0 / r) ** 6.0) / (self.r0 ** 2.0 * r)
        # pressure = flow * self.R + volume / self.C + self.peep_valve

        return {"volume": volume, "pressure": pressure}

    def step(self, action):
        self.target = self.waveform.at(self.time)
        reward = -jnp.abs(self.target - self.state["pressure"])
        self.state = self.dynamics(self.state, action)

        self.time += self.dt

        return self.observation, reward, False, {}



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy as np

from deluca.envs.core import Env


class Lung(Env):
    """
    Todos:
        - Plot incrementally
        - Save fig as fig.self
        - Modify close to delete fig and accumulated data
    """

    metadata = {"render.modes": ["human", "rgb_array"]}

    # scale parameter controls the relative size of the radius compared to base
    def render(self, mode="human", scale=1.0):
        screen_width = 600
        screen_height = 400

        # world_width = self.x_threshold * 2
        # scale = screen_width / world_width
        bottom_rect_y = 100  # TOP OF bottom_rect
        # polewidth = 10.0
        # polelen = scale * (2 * self.length)
        bottom_width = 250.0
        bottom_height = 5.0
        top_width = 250.0
        top_height = 5.0

        if self.viewer is None:
            from gym.envs.classic_control import rendering

            self.viewer = rendering.Viewer(screen_width, screen_height)
            # set up bottom rect
            l, r, t, b = -bottom_width / 2, bottom_width / 2, bottom_height / 2, -bottom_height / 2
            # axleoffset = cartheight / 4.0
            bottom = rendering.FilledPolygon([(l, b), (l, t), (r, t), (r, b)])
            self.bottom_trans = rendering.Transform()
            bottom.add_attr(self.bottom_trans)
            self.viewer.add_geom(bottom)

            # set up top rect
            l, r, t, b = 0, top_width, top_height / 2, -top_height / 2
            # axleoffset = cartheight / 4.0
            top = rendering.FilledPolygon([(l, b), (l, t), (r, t), (r, b)])
            self.top_trans = rendering.Transform(translation=(-bottom_width / 2, 0))
            top.add_attr(self.top_trans)
            top.add_attr(self.bottom_trans)
            self.viewer.add_geom(top)

            # set up balloon circle
            balloon = rendering.make_circle(radius=20, res=30, filled=False)
            self.balloon_trans = rendering.Transform()
            balloon.add_attr(self.balloon_trans)
            balloon.add_attr(self.bottom_trans)
            self.viewer.add_geom(balloon)

        if self.state is None:
            return None

        # translate bottom, which will automatically translate the top and balloon
        self.bottom_trans.set_translation(screen_width / 2.0, bottom_rect_y)

        # Edit the balloon and top
        if "pressure" in self.state.keys():
            pres = self.state["pressure"]
        elif "normalized_pressures" in self.state.keys():
            pres = self.pressure
        print("pres:" + str(pres))
        radius_polynomial = np.array(
            [0.01 * pres, -1, 0, 0, 0, 0, 0, 1]
        )  # https://en.wikipedia.org/wiki/Two-balloon_experiment
        possible_radii = np.roots(radius_polynomial)
        print("possible_radii:" + str(possible_radii))
        real_possible_radii = [r for r in possible_radii if (not np.iscomplex(r) and r > 0)]
        print("real_possible_radii:" + str(real_possible_radii))
        # assert(len(real_possible_radii) == 1)
        # radius = real_possible_radii[0]
        radius = max(real_possible_radii) * scale
        print("radius:" + str(radius))
        self.balloon_trans.set_translation(0.6 * bottom_width / 2.0, 0.1 * 20 * radius)
        self.balloon_trans.set_scale(0.1 * radius, 0.1 * radius)

        num = 0.1 * 20 * radius
        den = ((0.8 * bottom_width) ** 2 + (0.1 * 20 * radius) ** 2) ** 0.5
        top_theta = 2 * np.arcsin(num / den)
        self.top_trans.set_rotation(top_theta)

        # cartx = x[0] * scale + screen_width / 2.0  # MIDDLE OF CART
        # self.carttrans.set_translation(cartx, carty)
        # self.poletrans.set_rotation(-x[2])

        return self.viewer.render(return_rgb_array=mode == "rgb_array")



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""deluca.agents._gpc"""
from numbers import Real
from typing import Callable

import jax
import jax.numpy as jnp
import numpy as np
from jax import grad
from jax import jit

from deluca.agents._lqr import LQR
from deluca.agents.core import Agent


def quad_loss(x: jnp.ndarray, u: jnp.ndarray) -> Real:
    """
    Quadratic loss.

    Args:
        x (jnp.ndarray):
        u (jnp.ndarray):

    Returns:
        Real
    """
    return jnp.sum(x.T @ x + u.T @ u)


class GPC(Agent):
    def __init__(
        self,
        A: jnp.ndarray,
        B: jnp.ndarray,
        Q: jnp.ndarray = None,
        R: jnp.ndarray = None,
        K: jnp.ndarray = None,
        start_time: int = 0,
        cost_fn: Callable[[jnp.ndarray, jnp.ndarray], Real] = None,
        H: int = 3,
        HH: int = 2,
        lr_scale: Real = 0.005,
        decay: bool = True,
    ) -> None:
        """
        Description: Initialize the dynamics of the model.

        Args:
            A (jnp.ndarray): system dynamics
            B (jnp.ndarray): system dynamics
            Q (jnp.ndarray): cost matrices (i.e. cost = x^TQx + u^TRu)
            R (jnp.ndarray): cost matrices (i.e. cost = x^TQx + u^TRu)
            K (jnp.ndarray): Starting policy (optional). Defaults to LQR gain.
            start_time (int):
            cost_fn (Callable[[jnp.ndarray, jnp.ndarray], Real]):
            H (postive int): history of the controller
            HH (positive int): history of the system
            lr_scale (Real):
            lr_scale_decay (Real):
            decay (Real):
        """

        cost_fn = cost_fn or quad_loss

        d_state, d_action = B.shape  # State & Action Dimensions

        self.A, self.B = A, B  # System Dynamics

        self.t = 0  # Time Counter (for decaying learning rate)

        self.H, self.HH = H, HH

        self.lr_scale, self.decay = lr_scale, decay

        self.bias = 0

        # Model Parameters
        # initial linear policy / perturbation contributions / bias
        # TODO: need to address problem of LQR with jax.lax.scan
        self.K = K if K is not None else LQR(self.A, self.B, Q, R).K

        self.M = jnp.zeros((H, d_action, d_state))

        # Past H + HH noises ordered increasing in time
        self.noise_history = jnp.zeros((H + HH, d_state, 1))

        # past state and past action
        self.state, self.action = jnp.zeros((d_state, 1)), jnp.zeros((d_action, 1))

        def last_h_noises():
            """Get noise history"""
            return jax.lax.dynamic_slice_in_dim(self.noise_history, -H, H)

        self.last_h_noises = last_h_noises

        def policy_loss(M, w):
            """Surrogate cost function"""

            def action(state, h):
                """Action function"""
                return -self.K @ state + jnp.tensordot(
                    M, jax.lax.dynamic_slice_in_dim(w, h, H), axes=([0, 2], [0, 1])
                )

            def evolve(state, h):
                """Evolve function"""
                return self.A @ state + self.B @ action(state, h) + w[h + H], None

            final_state, _ = jax.lax.scan(evolve, np.zeros((d_state, 1)), np.arange(H - 1))
            return cost_fn(final_state, action(final_state, HH - 1))

        self.policy_loss = policy_loss
        self.grad = jit(grad(policy_loss, (0, 1)))

    def __call__(self, state: jnp.ndarray) -> jnp.ndarray:
        """
        Description: Return the action based on current state and internal parameters.

        Args:
            state (jnp.ndarray): current state

        Returns:
           jnp.ndarray: action to take
        """

        action = self.get_action(state)
        self.update(state, action)
        return action

    def update(self, state: jnp.ndarray, u: jnp.ndarray) -> None:
        """
        Description: update agent internal state.

        Args:
            state (jnp.ndarray):

        Returns:
            None
        """
        noise = state - self.A @ self.state - self.B @ u
        self.noise_history = jax.ops.index_update(self.noise_history, 0, noise)
        self.noise_history = jnp.roll(self.noise_history, -1, axis=0)

        delta_M, delta_bias = self.grad(self.M, self.noise_history)

        lr = self.lr_scale
        lr *= (1 / (self.t + 1)) if self.decay else 1
        self.M -= lr * delta_M
        self.bias -= lr * delta_bias

        # update state
        self.state = state

        self.t += 1

    def get_action(self, state: jnp.ndarray) -> jnp.ndarray:
        """
        Description: get action from state.

        Args:
            state (jnp.ndarray):

        Returns:
            jnp.ndarray
        """
        return -self.K @ state + jnp.tensordot(self.M, self.last_h_noises(), axes=([0, 2], [0, 1]))



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import inspect

import dill as pickle
import jax.numpy as jnp

from deluca.envs.lung import BreathWaveform
from deluca.envs.lung.core import Lung


class LearnedLung(Lung):
    def __init__(
        self,
        weights,
        pressure_mean=0.0,
        pressure_std=1.0,
        PEEP=5,
        input_dim=3,
        history_len=5,
        dt=0.03,
        waveform=None,
        reward_fn=None,
    ):
        self.viewer = None
        self.weights = weights
        self.pressure_mean = pressure_mean
        self.pressure_std = pressure_std
        self.input_dim = input_dim
        self.history_len = history_len
        self.dt = dt
        self.PEEP = PEEP
        self.waveform = waveform or BreathWaveform()
        self.reset()

    def reset(self):
        # pressure
        self.pressure = 0
        self.time = 0

        # NOTE: we don't combine this into a single trajectory because we must
        # roll these arrays at different times (e.g., u_ins, u_outs before
        # running through model, normalized pressures after)
        u_ins = jnp.zeros(self.history_len)
        u_outs = jnp.ones(self.history_len)
        normalized_pressures = (
            jnp.ones(self.history_len) * (self.PEEP - self.pressure_mean) / self.pressure_std
        )
        self.state = {
            "u_ins": u_ins,
            "u_outs": u_outs,
            "normalized_pressures": normalized_pressures,
        }

        return self.observation

    @classmethod
    def from_torch(cls, path):
        torch_sim = pickle.load(open(path, "rb"))
        args = {
            key: val
            for key, val in torch_sim.items()
            if key in inspect.signature(LearnedLung.__init__).parameters.keys()
        }
        sim = cls(**args)

        weights = []
        for weight in sim.weights:
            weights.append(jnp.array(weight))

        sim.weights = weights

        return sim

    @property
    def observation(self):
        return {
            "measured": self.pressure,
            "target": self.waveform.at(self.time),
            "dt": self.dt,
            "phase": self.waveform.phase(self.time),
        }

    def dynamics(self, state, action):
        """
        pressure: (u_in, u_out, normalized pressure) histories
        action: (u_in, u_out)
        """

        u_ins, u_outs, normalized_pressures = (
            state["u_ins"],
            state["u_outs"],
            state["normalized_pressures"],
        )
        u_in, u_out = action

        u_in /= 50.0
        u_out = u_out * 2.0 - 1.0

        u_ins = jnp.roll(u_ins, shift=-1)
        u_ins = u_ins.at[-1].set(u_in)
        u_outs = jnp.roll(u_outs, shift=-1)
        u_outs = u_outs.at[-1].set(u_out)

        normalized_pressure = jnp.concatenate((u_ins, u_outs, normalized_pressures))
        for i in range(0, len(self.weights), 2):
            normalized_pressure = self.weights[i] @ normalized_pressure + self.weights[i + 1]
            if i <= len(self.weights) - 4:
                normalized_pressure = jnp.tanh(normalized_pressure)

        normalized_pressures = jnp.roll(normalized_pressures, shift=-1)
        normalized_pressures = normalized_pressures.at[-1].set(normalized_pressure.squeeze())

        return {"u_ins": u_ins, "u_outs": u_outs, "normalized_pressures": normalized_pressures}

    def step(self, action):
        self.state = self.dynamics(self.state, action)
        self.pressure = (
            self.state["normalized_pressures"][-1] * self.pressure_std
        ) + self.pressure_mean
        self.pressure = jnp.clip(self.pressure, 0.0, 100.0)

        self.target = self.waveform.at(self.time)
        reward = -jnp.abs(self.target - self.pressure)

        self.time += self.dt

        return self.observation, reward, False, {}



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""deluca.agents._lqr"""
import jax.numpy as jnp
from scipy.linalg import solve_discrete_are as dare

from deluca.agents.core import Agent


class LQR(Agent):
    """
    LQR
    """

    # TODO: need to address problem of LQR with jax.lax.scan

    def __init__(
        self, A: jnp.ndarray, B: jnp.ndarray, Q: jnp.ndarray = None, R: jnp.ndarray = None
    ) -> None:
        """
        Description: Initialize the infinite-time horizon LQR.
        Args:
            A (jnp.ndarray): system dynamics
            B (jnp.ndarray): system dynamics
            Q (jnp.ndarray): cost matrices (i.e. cost = x^TQx + u^TRu)
            R (jnp.ndarray): cost matrices (i.e. cost = x^TQx + u^TRu)

        Returns:
            None
        """

        state_size, action_size = B.shape

        if Q is None:
            Q = jnp.identity(state_size, dtype=jnp.float32)

        if R is None:
            R = jnp.identity(action_size, dtype=jnp.float32)

        # solve the ricatti equation
        X = dare(A, B, Q, R)

        # compute LQR gain
        self.K = jnp.linalg.inv(B.T @ X @ B + R) @ (B.T @ X @ A)

    def __call__(self, state) -> jnp.ndarray:
        """
        Description: Return the action based on current state and internal parameters.

        Args:
            state (float/numpy.ndarray): current state

        Returns:
           jnp.ndarray: action to take
        """
        return -self.K @ state
